import time

import flask
import sepiida.environment
import sepiida.requests
import prometheus_client.core
import prometheus_client.exposition

REQUEST_LATENCY = prometheus_client.Histogram( # pylint: disable=unexpected-keyword-arg
    'request_latency_seconds', 'Flask Request Latency',
    ['service', 'method', 'endpoint'],
    namespace='sepiida',
)
REQUEST_COUNT = prometheus_client.Counter( # pylint: disable=unexpected-keyword-arg
    'flask_request_count', 'Flask Request Count',
    ['service', 'method', 'endpoint', 'http_status'],
    namespace='sepiida',
)
USER_SESSION_REQUEST_COUNT = prometheus_client.Counter( # pylint: disable=unexpected-keyword-arg
    'user_session_request_count', 'Requests Made By User Session',
    ['service', 'method', 'endpoint'],
    namespace='sepiida',
)
PRIVILEGED_SESSION_REQUEST_COUNT = prometheus_client.Counter( # pylint: disable=unexpected-keyword-arg
    'privileged_session_request_count', 'Requests Made by Privileged Session',
    ['service', 'method', 'endpoint'],
    namespace='sepiida',
)

def metrics():
    registry = prometheus_client.core.REGISTRY
    args = flask.request.args
    if 'name[]' in args:
        registry = registry.restricted_registry(args.getlist('name[]'))
    try:
        output = prometheus_client.exposition.generate_latest(registry)
    except Exception: # pylint: disable=broad-except
        return 500, 'error generating metric output', {}
    response = flask.make_response(output, 200)
    response.headers['Content-Type'] = prometheus_client.exposition.CONTENT_TYPE_LATEST
    return response

def before_request():
    flask.request.metrics = {
        'start'                         : time.time(),
    }

def after_request(response):
    # We may not have the metrics attribute if our response is generated by another before_request middleware
    if not hasattr(flask.request, 'metrics'):
        return response
    latency = time.time() - flask.request.metrics['start']
    settings = sepiida.environment.get()
    REQUEST_LATENCY.labels(
        settings.SERVER_NAME,
        flask.request.method,
        flask.request.path,
    ).observe(latency)
    REQUEST_COUNT.labels(
        settings.SERVER_NAME,
        flask.request.method,
        flask.request.path,
        response.status_code,
    ).inc()
    privileged_session_request_count = sepiida.requests.privileged_session_request_count()
    if privileged_session_request_count is not None:
        PRIVILEGED_SESSION_REQUEST_COUNT.labels( # pylint: disable=no-member
            settings.SERVER_NAME,
            flask.request.method,
            flask.request.path,
        ).inc(privileged_session_request_count)
    user_session_request_count = sepiida.requests.user_session_request_count()
    if user_session_request_count is not None:
        USER_SESSION_REQUEST_COUNT.labels( # pylint: disable=no-member
            settings.SERVER_NAME,
            flask.request.method,
            flask.request.path,
        ).inc(user_session_request_count)
    return response

def add_handler(app):
    app.route('/metrics/')(metrics)
    app.before_request(before_request)
    app.after_request(after_request)



import logging

import flask
import raven.base
import raven.utils.conf
from raven.contrib.flask import Sentry
from werkzeug.contrib.fixers import ProxyFix

import sepiida.environment

LOGGER = logging.getLogger(__name__)

def handle_reverse_proxy(app):
    LOGGER.info("Adding support for nginx reverse proxy fix in sentry")
    app.wsgi_app = ProxyFix(app.wsgi_app)


class SepiidaSentry(Sentry):
    def get_user_info(self, request):
        if hasattr(flask.g, 'current_user'):
            return flask.g.current_user
        return None


def add_sentry_support(application, sentry_dsn, version, tags=None):
    tags = tags or {}
    settings = sepiida.environment.get()
    options = raven.utils.conf.convert_options(application.config, defaults={
        'dsn'   : sentry_dsn,
        'include_paths' :
            set(application.config.get('SENTRY_INCLUDE_PATHS', [])) | set([application.import_name]),
        'extra' : {
            'app' : application,
        },
        # The above is taken from sentry's flask contrib module
        'environment'   : settings.ENVIRONMENT,
        'release'       : version,
        'site'          : settings.SERVER_NAME,
        'tags'          : tags,
    })

    client = raven.base.Client(**options)
    sentry = SepiidaSentry(app=application, client=client, dsn=sentry_dsn, logging=True, level=logging.ERROR)
    application.config.setdefault('RAVEN', sentry)
    logging.getLogger('webargs.flaskparser').setLevel(logging.CRITICAL)
    return sentry



import hashlib
import json as JSON
import logging
from datetime import datetime, timedelta

import jwt
from flask import g, request

import sepiida.environment
from sepiida.errors import APIException, Error
from sepiida.requests import privileged_session

LOGGER = logging.getLogger(__name__)

ALGORITHM = 'RS512'

def register_jwt_handlers(app, secret_key=None, secret_key_path=None):
    try:
        app.config['SEPIIDA_JWT_KEY'] = secret_key or read_secret_key(secret_key_path)
    except FileNotFoundError:
        raise Exception(
            "Unable to set up JWT handlers. You must either provide a SEPIIDA_JWT_KEY env var "
            "or put a key in /etc/authentise/authentise.jwt.key.pem")
    app.before_request(jwt_handler)

def jwt_handler():
    _jwt = request.headers.get('jwt')
    if _jwt:
        try:
            g.jwt = validate(_jwt)
        except Error as e:
            raise APIException(errors=[e])

def request_jwt():
    return g.get('jwt', None)

def validate(token, payload=None, audience=None):
    settings = sepiida.environment.get()
    iss = decode_iss(token)
    session = privileged_session()
    response = session.get(iss, timeout=5)
    if not response.ok:
        LOGGER.warning('Issuer (%s) request failed with status code %s: %s', iss, response.status_code, response.text)
        raise InvalidIssuerError("jwt-invalid-issuer", "Invalid issuer")

    response_json = response.json()
    if not response_json:
        raise InvalidIssuerError("jwt-invalid-issuer-response", "Invalid issuer response")

    secret = response_json.get('secret')
    if not secret:
        LOGGER.warning('Issuer (%s) request failed with missing secret token', iss)
        raise InvalidIssuerError("jwt-invalid-issuer-secret", "Invalid issuer secret")

    audience = audience or settings.SEPIIDA_JWT_AUD or request.url_root

    if not payload:
        payload = request.get_data()

    return decode(token, cert=secret, payload=payload, audience=audience)

def decode(token, cert=None, payload=None, verify=True, audience=None, issuer=None):
    options = {
        "require_exp": True,
        "require_iat": True,
        "require_nbf": True,
    }

    try:
        body = jwt.decode(token, cert, options=options, verify=False, audience=audience, issuer=issuer, leeway=10)
        if verify:
            jwt.decode(token, cert, options=options, verify=verify, audience=audience, issuer=issuer, leeway=10)
    except jwt.exceptions.MissingRequiredClaimError as ex:
        LOGGER.warning(ex)
        raise MissingClaimError("jwt-missing-claim", str(ex))
    except jwt.exceptions.InvalidIssuerError as ex:
        LOGGER.warning(ex)
        raise InvalidIssuerError("jwt-invalid-issuer", str(ex))
    except jwt.exceptions.InvalidAudienceError as ex:
        LOGGER.warning("InvalidAudience. Expected %s but received %s", audience, body['aud'])
        raise InvalidAudienceError("jwt-invalid-audience", str(ex))
    except jwt.exceptions.ExpiredSignatureError as ex:
        LOGGER.warning(ex)
        raise ExpiredSignatureError("jwt-signature-expired", str(ex))
    except jwt.exceptions.DecodeError as ex:
        LOGGER.warning(ex)
        raise DecodeError("jwt-decode-error", str(ex))

    if "exp" in body:
        body["exp"] = datetime.utcfromtimestamp(body["exp"])

    if "iat" in body:
        body["iat"] = datetime.utcfromtimestamp(body["iat"])

    if "nbf" in body:
        body["nbf"] = datetime.utcfromtimestamp(body["nbf"])

    if not verify:
        return body

    if "bdy" in body:
        bdy = body["bdy"]
        expected_bdy = _encode_payload(payload)
        if bdy != expected_bdy:
            LOGGER.warning('Invalid bdy claim: bdy "%s" does not match payload "%s"', bdy, expected_bdy)
            raise InvalidBodyError("jwt-invalid-body", 'Invalid Body')

    return body

def encode(key=None, **kwargs):
    utcnow = datetime.utcnow()
    settings = sepiida.environment.get()
    key = key if key else settings.SEPIIDA_JWT_KEY
    body = kwargs.copy()

    if "issuer" in body:
        body["iss"] = body.pop("issuer")

    if "subject" in body:
        body["sub"] = body.pop("subject")

    if "audience" in body:
        body["aud"] = body.pop("audience")

    if "exp" not in body:
        body["exp"] = utcnow + timedelta(seconds=900)

    if "iat" not in body:
        body["iat"] = utcnow

    if "nbf" not in body:
        body["nbf"] = utcnow

    payload = body.pop("payload", None)
    if payload:
        body["bdy"] = _encode_payload(payload)

    return jwt.encode(body, key, algorithm=ALGORITHM)

def decode_iss(token):
    body = decode(token, None, None, verify=False)
    iss = body.get("iss")
    if not iss:
        raise MissingClaimError("jwt-missing-cliam", 'Token is missing the "iss" claim')
    return iss

def read_secret_key(path=None):
    with open(path or "/etc/authentise/authentise.jwt.key.pem") as f:
        return f.read()

def _encode_payload(payload):
    if isinstance(payload, dict):
        payload = JSON.dumps(payload).encode()
    elif isinstance(payload, str):
        payload = payload.encode()
    return hashlib.sha256(payload).hexdigest()

class DecodeError(Error):
    pass

class InvalidClaimError(Error):
    pass

class MissingClaimError(Error):
    pass

class InvalidIssuerError(Error):
    pass

class InvalidAudienceError(Error):
    pass

class InvalidBodyError(Error):
    pass

class ExpiredSignatureError(Error):
    pass



'''
Tools to pull data out of the current session to re-use them
'''

import flask

import sepiida.session

def extract_credentials(_request):
    '''
    @param request url request object/dict
    @returns dict of {user,username, password, session-secret)
    dict value is None values missing from flask context'''
    user     = None
    username = None
    password = None
    session  = None

    if flask.has_request_context():
        user     = sepiida.session.current_user_uri()
        username = _request.authorization.get('username') if _request.authorization else None
        password = _request.authorization.get('password') if _request.authorization else None
        session  =  _request.cookies.get('session', None)

    return {
        'password'  : password,
        'session'   : session,
        'user'      : user,
        'username'  : username,
    }
